@@ -6,7 +6,7 @@ from abc import ABC
 import math
 import pprint
 import warnings
-from typing import Any, Dict, Iterable, List, Optional, Tuple
+from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple
 
 import numpy
 import pandas
@@ -17,6 +17,22 @@ import vtreat.transform
 import sklearn.base
 
 
+bad_sentinel = "_NA_"
+
+
+def replace_bad_with_sentinel(ar: List) -> numpy.ndarray:
+    """
+    Replace None/NaN entries in iterable with '_NA_'
+    :param ar: iterable
+    :return: one dimensional numpy array
+    """
+
+    ar = numpy.array(ar)
+    bads = pandas.isnull(ar)
+    ar[bads] = bad_sentinel
+    return ar
+
+
 def ready_data_frame(d) -> Tuple[pandas.DataFrame, type]:
     """
     Convert an array-like object to a data frame for processing.
@@ -33,7 +49,9 @@ def ready_data_frame(d) -> Tuple[pandas.DataFrame, type]:
     return d, orig_type
 
 
-def back_to_orig_type_data_frame(d: pandas.DataFrame, orig_type: type) -> Tuple[Any, List[str]]:
+def back_to_orig_type_data_frame(
+    d: pandas.DataFrame, orig_type: type
+) -> Tuple[Any, List[str]]:
     """
     Convert data frame back to ndarray if that was the original type.
 
@@ -55,11 +73,21 @@ class VarTransform(ABC):
     """
     Base class for vtreat transforms
     """
+
+    incoming_column_name_: str
+    derived_column_names_: List[str]
+    need_cross_treatment_: bool
+    treatment_: str
+    refitter_: Optional[Callable]
+    extra_args_: Optional[Dict[str, Any]]
+    params_: Optional[Dict[str, Any]]
+
     def __init__(
-            self,
-            incoming_column_name: str,
-            derived_column_names: Iterable[str],
-            treatment: str):
+        self,
+        incoming_column_name: str,
+        derived_column_names: Iterable[str],
+        treatment: str,
+    ):
         """
 
         :param incoming_column_name:
@@ -68,7 +96,7 @@ class VarTransform(ABC):
         """
 
         assert isinstance(incoming_column_name, str)
-        derived_column_names = [c for c in derived_column_names]
+        derived_column_names = list(derived_column_names)
         assert len(derived_column_names) > 0
         assert all([isinstance(dni, str) for dni in derived_column_names])
         assert treatment is not None
@@ -78,6 +106,8 @@ class VarTransform(ABC):
         self.treatment_ = treatment
         self.need_cross_treatment_ = False
         self.refitter_ = None
+        self.extra_args_ = None
+        self.params_ = None
 
     def transform(self, data_frame: pandas.DataFrame) -> pandas.DataFrame:
         """
@@ -90,15 +120,74 @@ class VarTransform(ABC):
 
         raise NotImplementedError("base method called")
 
+    def description_matrix(self) -> pandas.DataFrame:
+        """
+        Return description of transform as a data frame.
+
+        :return: description of transform.
+        """
+        raise NotImplementedError("base method called")
+
+
+class TreatmentPlan:
+    """
+    Class to carry treatment plans.
+    """
+
+    outcome_name: Optional[str]
+    cols_to_copy: Tuple[str, ...]
+    num_list: Tuple[str, ...]
+    cat_list: Tuple[str, ...]
+    xforms: Tuple[VarTransform, ...]
+
+    def __init__(
+            self,
+            *,
+            outcome_name: Optional[str] = None,
+            cols_to_copy: Optional[Iterable[str]] = None,
+            num_list: Optional[Iterable[str]] = None,
+            cat_list: Optional[Iterable[str]] = None,
+            xforms: Iterable[Optional[VarTransform]]):
+        self.outcome_name = outcome_name
+        if cols_to_copy is None:
+            self.cols_to_copy = tuple()
+        else:
+            assert not isinstance(cols_to_copy, str)
+            self.cols_to_copy = tuple(cols_to_copy)
+        if num_list is None:
+            self.num_list = tuple()
+        else:
+            assert not isinstance(num_list, str)
+            self.num_list = tuple(num_list)
+        if cat_list is None:
+            self.cat_list = tuple()
+        else:
+            assert not isinstance(cat_list, str)
+            self.cat_list = tuple(cat_list)
+        non_empty_xforms: List[VarTransform] = [x for x in xforms if x is not None]
+        self.xforms = tuple(non_empty_xforms)
+        for c in self.cols_to_copy:
+            assert isinstance(c, str)
+        for c in self.num_list:
+            assert isinstance(c, str)
+        for c in self.cat_list:
+            assert isinstance(c, str)
+        if len(self.xforms) < 1:
+            raise ValueError("no treatments generated")
+        for x in self.xforms:
+            assert isinstance(x, VarTransform)
+
 
 class MappedCodeTransform(VarTransform):
     """Class for transforms that are a dictionary mapping of values"""
+
     def __init__(
-            self,
-            incoming_column_name: str,
-            derived_column_name: str,
-            treatment: str,
-            code_book: pandas.DataFrame):
+        self,
+        incoming_column_name: str,
+        derived_column_name: str,
+        treatment: str,
+        code_book: pandas.DataFrame,
+    ):
         """
 
         :param incoming_column_name:
@@ -125,7 +214,7 @@ class MappedCodeTransform(VarTransform):
         derived_column_name = self.derived_column_names_[0]
         sf = pandas.DataFrame({incoming_column_name: data_frame[incoming_column_name]})
         bad_posns = vtreat.util.is_bad(sf[incoming_column_name])
-        sf.loc[bad_posns, incoming_column_name] = "_NA_"
+        sf.loc[bad_posns, incoming_column_name] = bad_sentinel
         res = pandas.merge(
             sf, self.code_book_, on=[self.incoming_column_name_], how="left", sort=False
         )  # ordered by left table rows
@@ -134,9 +223,31 @@ class MappedCodeTransform(VarTransform):
         res.loc[vtreat.util.is_bad(res[derived_column_name]), derived_column_name] = 0.0
         return res
 
+    def description_matrix(self) -> pandas.DataFrame:
+        """
+        Return description of transform as a data frame.
+
+        :return: description of transform.
+        """
+
+        description = pandas.DataFrame(
+            {
+                "treatment_class": "MappedCodeTransform",
+                "treatment": self.treatment_,
+                "orig_var": self.incoming_column_name_,
+                "variable": self.derived_column_names_[0],
+                "value": replace_bad_with_sentinel(
+                    self.code_book_[self.incoming_column_name_]
+                ),
+                "replacement": self.code_book_[self.derived_column_names_[0]].copy(),
+            }
+        )
+        return description
+
 
 class YAwareMappedCodeTransform(MappedCodeTransform):
     """Class for transforms that are a y-aware dictionary mapping of values"""
+
     def __init__(
         self,
         incoming_column_name: str,
@@ -173,6 +284,7 @@ class YAwareMappedCodeTransform(MappedCodeTransform):
 
 class CleanNumericTransform(VarTransform):
     """Class for numeric column cleaner."""
+
     def __init__(self, incoming_column_name: str, replacement_value: float):
         """
 
@@ -200,9 +312,29 @@ class CleanNumericTransform(VarTransform):
         res = pandas.DataFrame({self.derived_column_names_[0]: col})
         return res
 
+    def description_matrix(self) -> pandas.DataFrame:
+        """
+        Return description of transform as a data frame.
+
+        :return: description of transform.
+        """
+
+        description = pandas.DataFrame(
+            {
+                "treatment_class": ["CleanNumericTransform"],
+                "treatment": [self.treatment_],
+                "orig_var": [self.incoming_column_name_],
+                "variable": [self.derived_column_names_[0]],
+                "value": [bad_sentinel],
+                "replacement": [self.replacement_value_],
+            }
+        )
+        return description
+
 
 class IndicateMissingTransform(VarTransform):
     """Class for missing value indicator."""
+
     def __init__(self, incoming_column_name: str, derived_column_name: str):
         """
 
@@ -227,13 +359,32 @@ class IndicateMissingTransform(VarTransform):
         res = pandas.DataFrame({self.derived_column_names_[0]: col + 0.0})
         return res
 
+    def description_matrix(self) -> pandas.DataFrame:
+        """
+        Return description of transform as a data frame.
+
+        :return: description of transform.
+        """
+        description = pandas.DataFrame(
+            {
+                "treatment_class": ["IndicateMissingTransform"],
+                "treatment": [self.treatment_],
+                "orig_var": [self.incoming_column_name_],
+                "variable": [self.derived_column_names_[0]],
+                "value": [bad_sentinel],
+                "replacement": [1.0],
+            }
+        )
+        return description
+
 
 def fit_clean_code(
-        *,
-        incoming_column_name: str,
-        x,
-        params: Dict[str, Any],
-        imputation_map: Dict[str, Any]) -> Optional[VarTransform]:
+    *,
+    incoming_column_name: str,
+    x,
+    params: Dict[str, Any],
+    imputation_map: Dict[str, Any],
+) -> Optional[VarTransform]:
     """
     Fit numeric clean column imputation transform
 
@@ -280,11 +431,13 @@ def fit_clean_code(
 
 
 def fit_regression_impact_code(
-        *,
-        incoming_column_name: str,
-        x, y,
-        extra_args: Optional[Dict[str, Any]],
-        params: Dict[str, Any]) -> Optional[VarTransform]:
+    *,
+    incoming_column_name: str,
+    x,
+    y,
+    extra_args: Optional[Dict[str, Any]],
+    params: Dict[str, Any],
+) -> Optional[VarTransform]:
     """
     Fit regression impact code transform
 
@@ -318,11 +471,13 @@ def fit_regression_impact_code(
 
 
 def fit_regression_deviation_code(
-        *,
-        incoming_column_name: str,
-        x, y,
-        extra_args: Optional[Dict[str, Any]],
-        params: Dict[str, Any]) -> Optional[VarTransform]:
+    *,
+    incoming_column_name: str,
+    x,
+    y,
+    extra_args: Optional[Dict[str, Any]],
+    params: Dict[str, Any],
+) -> Optional[VarTransform]:
     """
     Fit regression deviation code transform
 
@@ -353,11 +508,13 @@ def fit_regression_deviation_code(
 
 
 def fit_binomial_impact_code(
-        *,
-        incoming_column_name: str,
-        x, y,
-        extra_args: Dict[str, Any],
-        params: Dict[str, Any]) -> Optional[VarTransform]:
+    *,
+    incoming_column_name: str,
+    x,
+    y,
+    extra_args: Dict[str, Any],
+    params: Dict[str, Any],
+) -> Optional[VarTransform]:
     """
     Fit categorical impact code.
 
@@ -368,7 +525,9 @@ def fit_binomial_impact_code(
     :param params: control parameter dictionary
     :return:
     """
-    outcome_target = (extra_args["outcome_target"],)  # TODO: document why this is a tuple
+    outcome_target = (
+        extra_args["outcome_target"],
+    )  # TODO: document why this is a tuple
     var_suffix = extra_args["var_suffix"]
     y = numpy.asarray(numpy.asarray(y) == outcome_target, dtype=float)
     sf = vtreat.util.grouped_by_x_statistics(x, y)
@@ -395,6 +554,7 @@ def fit_binomial_impact_code(
 
 class IndicatorCodeTransform(VarTransform):
     """Class for indicator codes"""
+
     def __init__(
         self,
         incoming_column_name: str,
@@ -428,7 +588,7 @@ class IndicatorCodeTransform(VarTransform):
         incoming_column_name = self.incoming_column_name_
         sf = pandas.DataFrame({incoming_column_name: data_frame[incoming_column_name]})
         bad_posns = vtreat.util.is_bad(sf[incoming_column_name])
-        sf.loc[bad_posns, incoming_column_name] = "_NA_"
+        sf.loc[bad_posns, incoming_column_name] = bad_sentinel
         col = sf[self.incoming_column_name_]
 
         def f(i: int):
@@ -446,9 +606,33 @@ class IndicatorCodeTransform(VarTransform):
                 res[self.derived_column_names_[ii]] = f(ii)
         return res
 
+    def description_matrix(self) -> pandas.DataFrame:
+        """
+        Return description of transform as a data frame.
+
+        :return: description of transform.
+        """
+
+        description = pandas.DataFrame(
+            {
+                "treatment_class": "IndicatorCodeTransform",
+                "treatment": self.treatment_,
+                "orig_var": self.incoming_column_name_,
+                "variable": self.derived_column_names_.copy(),
+                "value": replace_bad_with_sentinel(self.levels_),
+                "replacement": 1.0,
+            }
+        )
+        return description
+
 
 def fit_indicator_code(
-    *, incoming_column_name: str, x, min_fraction: float, sparse_indicators: bool = False
+    *,
+    incoming_column_name: str,
+    x,
+    min_fraction: float = 0.0,
+    max_levels: Optional[int] = None,
+    sparse_indicators: bool = False,
 ) -> Optional[VarTransform]:
     """
     Fit indicator codes
@@ -456,18 +640,30 @@ def fit_indicator_code(
     :param incoming_column_name:
     :param x: training explanatory variables
     :param min_fraction:
+    :param max_levels:
     :param sparse_indicators:
     :return:
     """
 
     sf = pandas.DataFrame({incoming_column_name: x})
     bad_posns = vtreat.util.is_bad(sf[incoming_column_name])
-    sf.loc[bad_posns, incoming_column_name] = "_NA_"
+    sf.loc[bad_posns, incoming_column_name] = bad_sentinel
     counts = sf[incoming_column_name].value_counts()
     n = sf.shape[0]
     counts = counts[counts > 0]
     counts = counts[counts >= min_fraction * n]  # no more than 1/min_fraction symbols
     levels = [str(v) for v in counts.index]
+    if (max_levels is not None) and (len(levels) > max_levels):
+        level_frame = pandas.DataFrame({"levels": levels, "counts": counts})
+        level_frame.sort_values(
+            by=["counts", "levels"],
+            ascending=[False, True],
+            inplace=True,
+            axis=0,
+            ignore_index=True,
+        )
+        value_vec = level_frame["levels"].values
+        levels = [value_vec[i] for i in range(max_levels)]
     if len(levels) < 1:
         return None
     return IndicatorCodeTransform(
@@ -489,7 +685,7 @@ def fit_prevalence_code(incoming_column_name: str, x) -> Optional[VarTransform]:
 
     sf = pandas.DataFrame({"x": x})
     bad_posns = vtreat.util.is_bad(sf["x"])
-    sf.loc[bad_posns, "x"] = "_NA_"
+    sf.loc[bad_posns, "x"] = bad_sentinel
     sf.reset_index(inplace=True, drop=True)
     n = sf.shape[0]
     sf["_ni"] = 1.0
@@ -508,10 +704,10 @@ def fit_prevalence_code(incoming_column_name: str, x) -> Optional[VarTransform]:
 
 # noinspection PyPep8Naming
 def _prepare_variable_lists(
-        *,
-        X,
-        cols_to_copy: Optional[Iterable[str]],
-        var_list: Optional[Iterable[str]],
+    *,
+    X,
+    cols_to_copy: Optional[Iterable[str]],
+    var_list: Optional[Iterable[str]],
 ) -> Tuple[List[str], List[str], List[str], List[str], List[str]]:
     """
     Prepare lists of variables for variable treatment.
@@ -522,16 +718,16 @@ def _prepare_variable_lists(
     :return: cat_list, cols_to_copy, mis_list, num_list, var_list lists
     """
     if var_list is None:
-        var_list = [co for co in X.columns]
+        var_list = list(X.columns)
     else:
-        var_list = [co for co in var_list]
+        var_list = list(var_list)
     if len(var_list) < 1:
-        var_list = [co for co in X.columns]
+        var_list = list(X.columns)
     assert len(var_list) > 0
     if cols_to_copy is None:
         cols_to_copy = []
     else:
-        cols_to_copy = [c for c in cols_to_copy]
+        cols_to_copy = list(cols_to_copy)
     copy_set = set(cols_to_copy)
     var_list = [co for co in var_list if (not (co in copy_set))]
     v_counts = {v: vtreat.util.get_unique_value_count(X[v]) for v in var_list}
@@ -539,8 +735,8 @@ def _prepare_variable_lists(
     if len(var_list) <= 0:
         raise ValueError("no variables")
     n = X.shape[0]
-    all_bad = []
-    mis_list = []
+    all_bad: List[str] = []
+    mis_list: List[str] = []
     for vi in var_list:
         n_bad = numpy.sum(vtreat.util.is_bad(X[vi]))
         if n_bad >= n:
@@ -563,14 +759,15 @@ def _prepare_variable_lists(
 
 # noinspection PyPep8Naming
 def fit_numeric_outcome_treatment(
-        *,
-        X, y,
-        var_list: Optional[Iterable[str]],
-        outcome_name: str,
-        cols_to_copy: Optional[Iterable[str]],
-        params: Dict[str, Any],
-        imputation_map: Dict[str, Any],
-):
+    *,
+    X,
+    y,
+    var_list: Optional[Iterable[str]],
+    outcome_name: str,
+    cols_to_copy: Optional[Iterable[str]],
+    params: Dict[str, Any],
+    imputation_map: Dict[str, Any],
+) -> TreatmentPlan:
     """
     Fit set of treatments in a regression situation.
 
@@ -584,14 +781,16 @@ def fit_numeric_outcome_treatment(
     :return: transform plan
     """
     cat_list, cols_to_copy, mis_list, num_list, var_list = _prepare_variable_lists(
-        X=X, cols_to_copy=cols_to_copy, var_list=var_list)
-    xforms = []
+        X=X, cols_to_copy=cols_to_copy, var_list=var_list
+    )
+    xforms: List[Optional[VarTransform]] = []
     if "missing_indicator" in params["coders"]:
         for vi in mis_list:
             xforms.append(
                 IndicateMissingTransform(
                     incoming_column_name=vi, derived_column_name=vi + "_is_bad"
-                ))
+                )
+            )
     if "clean_copy" in params["coders"]:
         for vi in num_list:
             xform = fit_clean_code(
@@ -601,11 +800,9 @@ def fit_numeric_outcome_treatment(
                 imputation_map=imputation_map,
             )
             if xform is not None:
-                # noinspection PyTypeChecker
                 xforms.append(xform)
     for vi in cat_list:
         if "impact_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_regression_impact_code(
                     incoming_column_name=vi,
@@ -616,7 +813,6 @@ def fit_numeric_outcome_treatment(
                 )
             )
         if "deviation_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_regression_deviation_code(
                     incoming_column_name=vi,
@@ -627,12 +823,10 @@ def fit_numeric_outcome_treatment(
                 )
             )
         if "prevalence_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
             )
         if "indicator_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_indicator_code(
                     incoming_column_name=vi,
@@ -641,29 +835,29 @@ def fit_numeric_outcome_treatment(
                     sparse_indicators=params["sparse_indicators"],
                 )
             )
-    xforms = [xf for xf in xforms if xf is not None]
     for stp in params["user_transforms"]:
         stp.fit(X=X[var_list], y=y)
-    return {
-        "outcome_name": outcome_name,
-        "cols_to_copy": cols_to_copy,
-        "num_list": num_list,
-        "cat_list": cat_list,
-        "xforms": xforms,
-    }
+    return TreatmentPlan(
+        outcome_name=outcome_name,
+        cols_to_copy=cols_to_copy,
+        num_list=num_list,
+        cat_list=cat_list,
+        xforms=xforms,
+    )
 
 
 # noinspection PyPep8Naming
 def fit_binomial_outcome_treatment(
     *,
-    X, y,
+    X,
+    y,
     outcome_target,
     var_list: Optional[Iterable[str]],
     outcome_name: str,
     cols_to_copy: Optional[Iterable[str]],
     params: Dict[str, Any],
     imputation_map: Dict[str, Any],
-):
+) -> TreatmentPlan:
     """
 
     :param X: training explanatory values
@@ -677,14 +871,16 @@ def fit_binomial_outcome_treatment(
     :return: transform plan
     """
     cat_list, cols_to_copy, mis_list, num_list, var_list = _prepare_variable_lists(
-        X=X, cols_to_copy=cols_to_copy, var_list=var_list)
-    xforms = []
+        X=X, cols_to_copy=cols_to_copy, var_list=var_list
+    )
+    xforms: List[Optional[VarTransform]] = []
     if "missing_indicator" in params["coders"]:
         for vi in mis_list:
             xforms.append(
                 IndicateMissingTransform(
                     incoming_column_name=vi, derived_column_name=vi + "_is_bad"
-                ))
+                )
+            )
     if "clean_copy" in params["coders"]:
         for vi in num_list:
             xform = fit_clean_code(
@@ -694,12 +890,10 @@ def fit_binomial_outcome_treatment(
                 imputation_map=imputation_map,
             )
             if xform is not None:
-                # noinspection PyTypeChecker
                 xforms.append(xform)
     extra_args = {"outcome_target": outcome_target, "var_suffix": ""}
     for vi in cat_list:
         if "logit_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_binomial_impact_code(
                     incoming_column_name=vi,
@@ -710,12 +904,10 @@ def fit_binomial_outcome_treatment(
                 )
             )
         if "prevalence_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
             )
         if "indicator_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_indicator_code(
                     incoming_column_name=vi,
@@ -724,28 +916,28 @@ def fit_binomial_outcome_treatment(
                     sparse_indicators=params["sparse_indicators"],
                 )
             )
-    xforms = [xf for xf in xforms if xf is not None]
     for stp in params["user_transforms"]:
         stp.fit(X=X[var_list], y=y)
-    return {
-        "outcome_name": outcome_name,
-        "cols_to_copy": cols_to_copy,
-        "num_list": num_list,
-        "cat_list": cat_list,
-        "xforms": xforms,
-    }
+    return TreatmentPlan(
+        outcome_name=outcome_name,
+        cols_to_copy=cols_to_copy,
+        num_list=num_list,
+        cat_list=cat_list,
+        xforms=xforms,
+    )
 
 
 # noinspection PyPep8Naming
 def fit_multinomial_outcome_treatment(
-        *,
-        X, y,
-        var_list: Optional[Iterable[str]],
-        outcome_name: str,
-        cols_to_copy: Optional[Iterable[str]],
-        params: Dict[str, Any],
-        imputation_map: Dict[str, Any],
-):
+    *,
+    X,
+    y,
+    var_list: Optional[Iterable[str]],
+    outcome_name: str,
+    cols_to_copy: Optional[Iterable[str]],
+    params: Dict[str, Any],
+    imputation_map: Dict[str, Any],
+) -> TreatmentPlan:
     """
     Fit a variable treatment for multinomial outcomes.
 
@@ -762,14 +954,16 @@ def fit_multinomial_outcome_treatment(
     outcomes = [oi for oi in set(y)]
     assert len(outcomes) > 1
     cat_list, cols_to_copy, mis_list, num_list, var_list = _prepare_variable_lists(
-        X=X, cols_to_copy=cols_to_copy, var_list=var_list)
-    xforms = []
+        X=X, cols_to_copy=cols_to_copy, var_list=var_list
+    )
+    xforms: List[Optional[VarTransform]] = []
     if "missing_indicator" in params["coders"]:
         for vi in mis_list:
             xforms.append(
                 IndicateMissingTransform(
                     incoming_column_name=vi, derived_column_name=vi + "_is_bad"
-                ))
+                )
+            )
     if "clean_copy" in params["coders"]:
         for vi in num_list:
             xform = fit_clean_code(
@@ -779,7 +973,6 @@ def fit_multinomial_outcome_treatment(
                 imputation_map=imputation_map,
             )
             if xform is not None:
-                # noinspection PyTypeChecker
                 xforms.append(xform)
     for vi in cat_list:
         for outcome in outcomes:
@@ -788,7 +981,6 @@ def fit_multinomial_outcome_treatment(
                     "outcome_target": outcome,
                     "var_suffix": ("_" + str(outcome)),
                 }
-                # noinspection PyTypeChecker
                 xforms.append(
                     fit_binomial_impact_code(
                         incoming_column_name=vi,
@@ -799,12 +991,10 @@ def fit_multinomial_outcome_treatment(
                     )
                 )
         if "prevalence_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
             )
         if "indicator_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_indicator_code(
                     incoming_column_name=vi,
@@ -813,30 +1003,27 @@ def fit_multinomial_outcome_treatment(
                     sparse_indicators=params["sparse_indicators"],
                 )
             )
-    xforms = [xf for xf in xforms if xf is not None]
-    if len(xforms) <= 0:
-        raise ValueError("no variables created")
     for stp in params["user_transforms"]:
         stp.fit(X=X[var_list], y=y)
-    return {
-        "outcome_name": outcome_name,
-        "cols_to_copy": cols_to_copy,
-        "num_list": num_list,
-        "cat_list": cat_list,
-        "xforms": xforms,
-    }
+    return TreatmentPlan(
+        outcome_name=outcome_name,
+        cols_to_copy=cols_to_copy,
+        num_list=num_list,
+        cat_list=cat_list,
+        xforms=xforms,
+    )
 
 
 # noinspection PyPep8Naming
 def fit_unsupervised_treatment(
-        *,
-        X,
-        var_list: Optional[Iterable[str]],
-        outcome_name: str,
-        cols_to_copy: Optional[Iterable[str]],
-        params: Dict[str, Any],
-        imputation_map: Dict[str, Any],
-):
+    *,
+    X,
+    var_list: Optional[Iterable[str]],
+    outcome_name: str,
+    cols_to_copy: Optional[Iterable[str]],
+    params: Dict[str, Any],
+    imputation_map: Dict[str, Any],
+) -> TreatmentPlan:
     """
     Fit a data treatment in the unsupervised case.
 
@@ -850,14 +1037,16 @@ def fit_unsupervised_treatment(
     """
 
     cat_list, cols_to_copy, mis_list, num_list, var_list = _prepare_variable_lists(
-        X=X, cols_to_copy=cols_to_copy, var_list=var_list)
-    xforms = []
+        X=X, cols_to_copy=cols_to_copy, var_list=var_list
+    )
+    xforms: List[Optional[VarTransform]] = []
     if "missing_indicator" in params["coders"]:
         for vi in mis_list:
             xforms.append(
                 IndicateMissingTransform(
                     incoming_column_name=vi, derived_column_name=vi + "_is_bad"
-                ))
+                )
+            )
     if "clean_copy" in params["coders"]:
         for vi in num_list:
             xform = fit_clean_code(
@@ -867,42 +1056,40 @@ def fit_unsupervised_treatment(
                 imputation_map=imputation_map,
             )
             if xform is not None:
-                # noinspection PyTypeChecker
                 xforms.append(xform)
     for vi in cat_list:
         if "prevalence_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
             )
         if "indicator_code" in params["coders"]:
-            # noinspection PyTypeChecker
             xforms.append(
                 fit_indicator_code(
                     incoming_column_name=vi,
                     x=numpy.asarray(X[vi]),
                     min_fraction=params["indicator_min_fraction"],
+                    max_levels=params["indicator_max_levels"],
                     sparse_indicators=params["sparse_indicators"],
                 )
             )
-    xforms = [xf for xf in xforms if xf is not None]
     for stp in params["user_transforms"]:
         stp.fit(X=X[var_list], y=None)
-    return {
-        "outcome_name": outcome_name,
-        "cols_to_copy": cols_to_copy,
-        "num_list": num_list,
-        "cat_list": cat_list,
-        "xforms": xforms,
-    }
+    return TreatmentPlan(
+        outcome_name=outcome_name,
+        cols_to_copy=cols_to_copy,
+        num_list=num_list,
+        cat_list=cat_list,
+        xforms=xforms,
+    )
 
 
 def pre_prep_frame(
-        x: pandas.DataFrame,
-        *,
-        col_list: Optional[Iterable[str]],
-        cols_to_copy: Optional[Iterable[str]],
-        cat_cols: Optional[Iterable[str]] = None) -> pandas.DataFrame:
+    x: pandas.DataFrame,
+    *,
+    col_list: Optional[Iterable[str]],
+    cols_to_copy: Optional[Iterable[str]],
+    cat_cols: Optional[Iterable[str]] = None,
+) -> pandas.DataFrame:
     """
     Create a copy of pandas.DataFrame x restricted to col_list union cols_to_copy with col_list - cols_to_copy
     converted to only string and numeric types.  New pandas.DataFrame has trivial indexing.  If col_list
@@ -918,13 +1105,13 @@ def pre_prep_frame(
     if cols_to_copy is None:
         cols_to_copy = []
     else:
-        cols_to_copy = [c for c in cols_to_copy]
+        cols_to_copy = list(cols_to_copy)
     if col_list is None:
         col_list = []
-    if len(col_list) <= 0:
-        col_list = [co for co in x.columns]
     else:
-        col_list = [c for c in col_list]
+        col_list = list(col_list)
+    if len(col_list) <= 0:
+        col_list = list(x.columns)
     x_set = set(x.columns)
     col_set = set(col_list)
     for ci in cols_to_copy:
@@ -986,12 +1173,8 @@ def _mean_of_single_column_pandas_list(val_list: Iterable[pandas.DataFrame]) ->
 
 
 def cross_patch_refit_y_aware_cols(
-        *,
-        x: pandas.DataFrame,
-        y,
-        res: pandas.DataFrame,
-        plan,
-        cross_plan) -> None:
+    *, x: pandas.DataFrame, y, res: pandas.DataFrame, plan: TreatmentPlan, cross_plan
+) -> None:
     """
     Re fit the y-aware columns according to cross plan.
     Clears out refitter_ values to None.
@@ -1006,12 +1189,12 @@ def cross_patch_refit_y_aware_cols(
     """
 
     if cross_plan is None or len(cross_plan) <= 1:
-        for xf in plan["xforms"]:
+        for xf in plan.xforms:
             xf.refitter_ = None
         return
     incoming_colset = set(x.columns)
     derived_colset = set(res.columns)
-    for xf in plan["xforms"]:
+    for xf in plan.xforms:
         if not xf.need_cross_treatment_:
             continue
         incoming_column_name = xf.incoming_column_name_
@@ -1065,17 +1248,13 @@ def cross_patch_refit_y_aware_cols(
                 pi[derived_column_name]
             ).reshape((len(pi),))
         res.loc[vtreat.util.is_bad(res[derived_column_name]), derived_column_name] = avg
-    for xf in plan["xforms"]:
+    for xf in plan.xforms:
         xf.refitter_ = None
 
 
 def cross_patch_user_y_aware_cols(
-        *,
-        x: pandas.DataFrame,
-        y,
-        res: pandas.DataFrame,
-        params: Dict[str, Any],
-        cross_plan) -> None:
+    *, x: pandas.DataFrame, y, res: pandas.DataFrame, params: Dict[str, Any], cross_plan
+) -> None:
     """
     Re fit the user y-aware columns according to cross plan.
     Assumes each y-aware variable produces one derived column.
@@ -1129,19 +1308,19 @@ def cross_patch_user_y_aware_cols(
 
 
 def score_plan_variables(
-        cross_frame: pandas.DataFrame,
-        outcome,
-        plan,
-        params: Dict[str, Any],
-        *,
-        is_classification: bool = False
+    cross_frame: pandas.DataFrame,
+    outcome,
+    plan: TreatmentPlan,
+    params: Dict[str, Any],
+    *,
+    is_classification: bool = False,
 ) -> pandas.DataFrame:
     """
     Quality score variables to build up score frame.
 
     :param cross_frame: cross transformed explanatory variables
     :param outcome: dependent variable
-    :param plan: treatment plan dictionary
+    :param plan: treatment plan
     :param params: control parameter dictionary
     :param is_classification: logical, if True classification if False regression
     :return: score frame
@@ -1165,7 +1344,7 @@ def score_plan_variables(
         return description
 
     var_table = pandas.concat(
-        [describe_xf(xf) for xf in plan["xforms"]]
+        [describe_xf(xf) for xf in plan.xforms]
         + [
             describe_ut(ut)
             for ut in params["user_transforms"]
@@ -1210,15 +1389,13 @@ def score_plan_variables(
 
 
 def pseudo_score_plan_variables(
-        *,
-        cross_frame,
-        plan,
-        params: Dict[str, Any]) -> pandas.DataFrame:
+    *, cross_frame, plan: TreatmentPlan, params: Dict[str, Any]
+) -> pandas.DataFrame:
     """
     Build a score frame look-alike for unsupervised case.
 
     :param cross_frame: cross transformed explanatory variables
-    :param plan:
+    :param plan: treatment plan
     :param params: control parameter dictionary
     :return: score frame
     """
@@ -1241,7 +1418,7 @@ def pseudo_score_plan_variables(
         return description
 
     score_frame = pandas.concat(
-        [describe_xf(xf) for xf in plan["xforms"]]
+        [describe_xf(xf) for xf in plan.xforms]
         + [
             describe_ut(ut)
             for ut in params["user_transforms"]
@@ -1268,14 +1445,29 @@ class VariableTreatment(ABC, sklearn.base.BaseEstimator, sklearn.base.Transforme
     Class for variable treatments, implements much of the sklearn pipeline/transformer
     API. https://sklearn-template.readthedocs.io/en/latest/user_guide.html#transformer
     """
+
+    result_restriction: Optional[Set[str]]
+    outcome_name_: Optional[str]
+    outcome_target_: Optional[Any]
+    var_list_: List[str]
+    cols_to_copy_: List[str]
+    params_: Dict[str, Any]
+    plan_: Optional[TreatmentPlan]
+    score_frame_: Optional[pandas.DataFrame]
+    imputation_map_: Dict[str, Callable]
+    last_fit_x_id_: Optional[str]
+    cross_plan_: Optional[List[Dict[str, List[int]]]]
+    cross_rows_: Optional[int]
+    last_result_columns: Optional[List[str]]
+
     def __init__(
         self,
         *,
         var_list: Optional[Iterable[str]] = None,
         outcome_name: Optional[str] = None,
-        outcome_target=None,
+        outcome_target: Optional[Any] = None,
         cols_to_copy: Optional[Iterable[str]] = None,
-        params: Dict[str, Any] = None,
+        params: Optional[Dict[str, Any]] = None,
         imputation_map: Optional[Dict[str, Any]] = None,
     ):
         """
@@ -1308,7 +1500,10 @@ class VariableTreatment(ABC, sklearn.base.BaseEstimator, sklearn.base.Transforme
         self.outcome_target_ = outcome_target
         self.var_list_ = [vi for vi in var_list if vi not in set(cols_to_copy)]
         self.cols_to_copy_ = cols_to_copy
-        self.params_ = params.copy()
+        if params is not None:
+            self.params_ = params.copy()
+        else:
+            self.params_ = dict()
         self.imputation_map_ = imputation_map.copy()
         self.plan_ = None
         self.score_frame_ = None
@@ -1546,12 +1741,29 @@ class VariableTreatment(ABC, sklearn.base.BaseEstimator, sklearn.base.Transforme
         new_vars = new_vars + self.cols_to_copy_
         return new_vars
 
+    def description_matrix(self) -> pandas.DataFrame:
+        """
+        Return description of transform as a data frame. Does not encode user steps. Not yet implemented for
+        multinomial dependent variables.
+
+        :return: description of transform.
+        """
+
+        assert self.plan_ is not None  # type hint
+        assert self.score_frame_ is not None  # type hint
+        xform_steps = [xfi for xfi in self.plan_.xforms]
+        frames = [xfi.description_matrix() for xfi in xform_steps]
+        res = pandas.concat(frames).reset_index(inplace=False, drop=True)
+        # restrict down to non-constant variables
+        scored_vars = set(self.score_frame_["variable"][self.score_frame_["has_range"]])
+        usable = [v in scored_vars for v in res["variable"]]
+        res = res.loc[usable, :].reset_index(inplace=False, drop=True)
+        return res
+
 
 def perform_transform(
-        *,
-        x: pandas.DataFrame,
-        transform: VariableTreatment,
-        params: Dict[str, Any]) -> pandas.DataFrame:
+    *, x: pandas.DataFrame, transform: VariableTreatment, params: Dict[str, Any]
+) -> pandas.DataFrame:
     """
     Transform a data frame.
 
@@ -1560,8 +1772,8 @@ def perform_transform(
     :param params: control parameter dictionary
     :return: new data frame
     """
-    plan = transform.plan_
-    xform_steps = [xfi for xfi in plan["xforms"]]
+    assert transform.plan_ is not None  # type hint
+    xform_steps = [xfi for xfi in transform.plan_.xforms]
     user_steps = [stp for stp in params["user_transforms"]]
     # restrict down to to results we are going to use
     if (transform.result_restriction is not None) and (
@@ -1598,7 +1810,7 @@ def perform_transform(
     new_frames = [xfi.transform(x) for xfi in (xform_steps + user_steps)]
     new_frames = [frm for frm in new_frames if (frm is not None) and (frm.shape[1] > 0)]
     # see if we want to copy over any columns
-    copy_set = set(plan["cols_to_copy"])
+    copy_set = set(transform.plan_.cols_to_copy)
     to_copy = [ci for ci in x.columns if ci in copy_set]
     if len(to_copy) > 0:
         cp = x.loc[:, to_copy].copy()
@@ -1611,9 +1823,8 @@ def perform_transform(
 
 
 def limit_to_appropriate_columns(
-        *,
-        res: pandas.DataFrame,
-        transform: VariableTreatment) -> pandas.DataFrame:
+    *, res: pandas.DataFrame, transform: VariableTreatment
+) -> pandas.DataFrame:
     """
     Limit down to appropriate columns.
 
@@ -1621,8 +1832,9 @@ def limit_to_appropriate_columns(
     :param transform:
     :return:
     """
-    plan = transform.plan_
-    to_copy = set(plan["cols_to_copy"])
+    assert transform.plan_ is not None  # type hint
+    assert transform.score_frame_ is not None  # type hint
+    to_copy = set(transform.plan_.cols_to_copy)
     to_take = set(
         [
             ci
